package nl.arttech.bouncing;
import java.awt.Color;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Vector;

/**
 * OpenSoundControl message class, based on original classes by Ben Chun
 *
 * @author  Ben Chun        	ben@benchun.net
 * @author  Mattijs Kneppers 	mattijs@arttech.nl
 * @version 1.0
 */

public class OscMessage {
	public String address;
    public Vector<Character> types = new Vector<Character>();
    public Vector<Object> arguments = new Vector<Object>();

    /**
     * Constructor for the OscMessage.
     *
     * @param   name    the message name
    */
    public OscMessage(String address) {
		this.address = address;
    }

    /**
     * Adds a type/argument pair to the list of arguments
     *
     * @param   type           the argument data type
     * @param   argument       the argument value
    */
    public void addArg(Character type, Object argument) {
		types.addElement(type);
		arguments.addElement(argument);
    }

    /**
     * Directly sets the type and arg Vectors
     *
     * @param   types    a list of types
     * @param   args     a list of arguments matching the types
    */
    void setTypesAndArgs(Vector<Character> types, Vector<Object> args) {
		this.types = types;
		arguments = args;
    }

    /**
     * Returns a byte array representation of this message.
     *
     */
	byte[] getByteArray() throws IOException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		DataOutputStream stream = new DataOutputStream(baos);
		char type;

		// address (name)
		stream.writeBytes( address );
		alignStream( baos );

		// type tags
		stream.writeByte( ',' );  // comma indicates type tags
		Enumeration<Character> t = types.elements();
		while ( t.hasMoreElements() ) {
			type = t.nextElement().charValue();
			stream.writeByte( type );
		}
		alignStream(baos);

		// values
		t = types.elements();
		Enumeration<Object> a = arguments.elements();
		while ( t.hasMoreElements() ) {
			type = t.nextElement().charValue();
			switch(type) {
				case 'i':
					stream.writeInt(((Integer)a.nextElement()).intValue());
					break;
				case 'r':
					Color color = (Color)a.nextElement();
					
					byte[] colorBytes = new byte[] {
			                (byte)(color.getRed()),
			                (byte)(color.getGreen()),
			                (byte)(color.getBlue()),
			                (byte)color.getAlpha()};
					
					stream.write(colorBytes);
					break;
				case 'f':
					stream.writeFloat(((Float)a.nextElement()).floatValue());
					break;
				case 'h':
					stream.writeLong(((Long)a.nextElement()).longValue());
					break;
				case 'd':
					stream.writeDouble(((Double)a.nextElement()).doubleValue());
					break;
				case 's':
					stream.writeBytes((String)a.nextElement());
					alignStream( baos );
					break;
				case 'T':
				case 'F':
				case 'N':
				case 'I':
				default:
			}
		}
		return baos.toByteArray();
    }

    /**
     * Make the stream end on a 4-byte boundary by padding it with
     * null characters.
     *
     * @param stream The stream to align.
     */
    private void alignStream(ByteArrayOutputStream stream) throws IOException {
        int pad = 4 - ( stream.size() % 4 );
        for (int i = 0; i < pad; i++)
            stream.write(0);
    }

    /**
     * Returns a string representation of this message
     *
     */
    @Override
	public String toString() {
		String temp = new String();
		temp += "Name: " + address + "\n";
		if (types.size() != arguments.size()) {
			temp += "Amount of types and amount of arguents do not match";
		}
		else {
			for (int i=0;i<arguments.size();i++) {
				temp += "Parameter " + i + ": Type " + (types.elementAt(i)).toString() + ", Arg ";
				switch((types.elementAt(i)).charValue()) {
					case 'i':
						temp += ((Integer)arguments.elementAt(i)).toString();
						break;
					case 'f':
						temp += ((Float)arguments.elementAt(i)).toString();
						break;
					case 'h':
						temp += ((Long)arguments.elementAt(i)).toString();
						break;
					case 'd':
						temp += ((Double)arguments.elementAt(i)).toString();
						break;
					case 's':
						temp += ((String)arguments.elementAt(i)).toString();
						break;
					case 'T':
					case 'F':
					case 'N':
					case 'I':
					default:
				}
				temp += "\n";
			}
		}
		return temp;
	}
}
