package nl.arttech.bouncing;
import java.awt.Color;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Vector;


/**
 * OpenSoundControl packet class. Based on original class by Ben Chun
 *
 * @author  Ben Chun        	ben@benchun.net
 * @author	Mattijs Kneppers	mattijs@arttech.nl
 * @version 1.0
 */

public class OscPacket {
	public static final byte[] timeZeroBytes = {0, 0, 0, 0, 0, 0, 0, 1};
	
	public byte[] ntpTimeTag = new byte[8]; 
    public Vector<OscMessage> messages = null;
    public InetAddress address = null;
    public int port = 0;

    /**
     * Constructor for all incoming packets and those outgoing packets
     * whose time, address, and port will be set later.
     *
    */
    public OscPacket() {
    	ntpTimeTag = timeZeroBytes;
		messages = new Vector<OscMessage>();
    }
    
    public OscPacket(DatagramPacket datagramPacket) {
		messages = new Vector<OscMessage>();
		ntpTimeTag = timeZeroBytes;
		port = datagramPacket.getPort();
		address = datagramPacket.getAddress();
		
		parseOscPacket(datagramPacket.getData(), datagramPacket.getLength(), this);
    }

    /**
     * Adds a message to this packet.
     *
     * @param   message   the message to add
    */
    public void addMessage(OscMessage message) {
		messages.addElement(message);
    }

    /**
     * Returns a byte array representation of this packet, suitable for
     * sending to OSC client applications.
     * 
     * Note: always makes bundles, no support for single OscMessages
     *
     */
	public byte[] getByteArray() throws IOException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		DataOutputStream stream = new DataOutputStream(baos);

		baos.write( ("#bundle").getBytes() );
		baos.write(0);
		
		stream.write(ntpTimeTag);

		// messages
		byte[] byteArray;
		for (OscMessage m : messages) {
			byteArray = m.getByteArray();
			stream.writeInt(byteArray.length); // bundles have message size tags
			baos.write(byteArray);
		}
		return baos.toByteArray();
    }	
	
    /**
     * Returns a string representation of this packet
     *
     */
    @Override
	public String toString() {
		String temp = new String();
		temp += "__OSC Packet__\n";
		if (address != null) {
			temp += "Address: " + address.toString() + "\n";
		}
		else {
			temp += "Address not set\n";
		}
		temp += "Port: " + port + "\n";
		
		temp += "Time: " + timeFormat(OscPacket.timestampToMs(ntpTimeTag)) + "\n";

		for (int i=0;i<messages.size();i++) {
			temp += "__Message " + i + "__\n" + (messages.elementAt(i)).toString();
		}
		return temp;
	}
       
	/**
	 * Converts an unsigned byte to a short.  By default, Java assumes that
	 * a byte is signed.
	 */
	static short unsignedByteToShort(byte b) {
		if((b & 0x80)==0x80) return (short) (128 + (b & 0x7f));
		else return (short) b;
	}

	/**
	 * Converts the timestamp (64-bit NTP format, number of seconds since 00:00 1-Jan-1900) to
	 * the amount of milliseconds since 00:00 1-Jan-1970 (UTC)
	 */
	public static long timestampToMs(byte[] timestamp) {
		if (timestamp.equals(OscPacket.timeZeroBytes)) return 0;

		double value = 0; 
		for (int i = 0; i < 8; i++) {
			double base = Math.pow(2, (3-i)*8);
			
			value += OscPacket.unsignedByteToShort(timestamp[i]) * base;
		}
		
		if (value == 0) return 0;
		
		// timestamp is relative to 1900, UTC is used by Java and is relative to 1970 
		double utc = value - (2208988800.0);
		
		// milliseconds
		return (long) (utc * 1000.0);
	}

	/**
	 * Converts the amount of milliseconds since 00:00 1-Jan-1970 (UTC) to 
	 * a timestamp (64-bit NTP format, number of seconds since 00:00 1-Jan-1900)
	 */	
	public static byte[] msToTimestamp(long time) {
		if (time == 0) {
			return OscPacket.timeZeroBytes;	
		}
	
		byte[] array = new byte[8];
	
		// NTP timestamp is relative to 1900, UTC is used by Java and is relative to 1970 
		double ntpTime = time / 1000. + (2208988800.0);		
		//Utils.post("ntpTime: " + timeFormat((long)ntpTime));
		
		// Converts a double into a 64-bit fixed point
		for (int i=0; i<8; i++) {
			// 2^24, 2^16, 2^8, .. 2^-32
			double base = Math.pow(2, (3-i)*8);
			
			// Capture byte value
			array[i] = (byte) (ntpTime / base);
			// Subtract captured value from remaining total
			ntpTime = ntpTime - (OscPacket.unsignedByteToShort(array[i]) * base);
		}
		
		// From RFC 2030: It is advisable to fill the non-significant
		// low order bits of the timestamp with a random, unbiased
		// bitstring, both to avoid systematic roundoff errors and as
		// a means of loop detection and replay detection.
		array[7] = (byte) (Math.random()*255.0);		
		
		return array;
	}

	/**
	 * OpenSoundControl methods
	 *
	 * Based on CommServer by Derek Clayton and dumpOSC by Matt Wright.
	 *
	 * Thanks to Jesse Gilbert (j@resistantstrain.net) for helping me with
	 * the static socket.
	 * 
	 * Ben
	 *
	 * Modifications for supporting NTP time stamps by Mattijs Kneppers
	 *
	 * @author  Ben Chun        	ben@benchun.net
	 * @author  Mattijs Kneppers	mattijs@arttech.nl
	 * @version 1.0
	 */
    /**
     * This method is based on dumpOSC::ParseOSCPacket.  It verifies
     * that the packet is well-formed and puts the data into an
     * OscPacket object, or else prints useful error messages about
     * what went wrong.
     *
     * @param   datagram  a byte array containing the OSC packet
     * @param   n         size of the byte array
     * @param   packet    the OscPacket object to put data into
     */

	@SuppressWarnings("unchecked")
	static String parseOscPacket(byte[] datagram, int n, OscPacket packet) {
		int size, i;

		if ( (n % 4) != 0 ) {
			return "SynthControl packet size (" + n + ") not a multiple of 4 bytes, dropped it.";
		}

		String dataString = new String(datagram);

		if ( ( n >= 8 ) && dataString.startsWith( "#time" ) ) {
			
			/* This is a time syc request. */
			System.out.println("Received time sync request.");
		}
		else if ( ( n >= 8 ) && dataString.startsWith( "#bundle" ) ) {
//			System.out.println("this is a bundle message");
			/* This is a bundle message. */

			if ( n < 16 ) {
				return "Bundle message too small (" + n + " bytes) for time tag, dropped it.";
			}
			
			/* Get the time tag (ben's old version) */
			Long time = new Long( Utils.toLong( Utils.copyBytes(datagram, 8, 8) ) );
//			System.out.println("found long time: " + time);
					
			if (time != 1) {
				/* Get the time tag (new version) */
				packet.ntpTimeTag = decodeTimestamp(datagram, 8);	// in seconds
//				System.out.println("found time tag: " + timestampToMs(packet.ntpTimeTag));
			}
			
			i = 16; /* Proceed from "#bundle\0" and time tags */
			while( i < n ) {
				size = ( Utils.toInt( Utils.copyBytes(datagram, i, 4 ) ) );
				if ((size % 4) != 0) {
					return "Bad size count" + size + "in bundle (not a multiple of 4)";
				}
				if ( (size + i + 4) > n ) {
					return "Bad size count" + size + "in bundle" + "(only" + (n-i-4) + "bytes left in entire bundle)";
				}

				/* Recursively handle element of bundle */
				byte[] remaining =  Utils.copyBytes(datagram, i+4, datagram.length - (i+4));
				parseOscPacket(remaining, size, packet);
				i += (4 + size);
			}
		}
		else {
//			System.out.println("this is NOT a bundle message");
			
			/* This is not a bundle message */
			Vector<Object> nameAndData = getStringAndData(datagram, n);

			String name = (String) nameAndData.firstElement();
			OscMessage message = new OscMessage(name);

			byte[] data = (byte[]) nameAndData.lastElement();
			@SuppressWarnings("rawtypes")
			Vector[] typesAndArgs = getTypesAndArgs(data);
			message.setTypesAndArgs(typesAndArgs[0], typesAndArgs[1]);

			packet.addMessage(message);
		}
		return "";
    }
    
	/**
	 * Will read 8 bytes of a message beginning at <code>pointer</code>
	 * and return it as a double, according to the NTP 64-bit timestamp
	 * format.
	 */
	static byte[] decodeTimestamp(byte[] array, int pointer) {
		byte[] timestamp = new byte[8];
		
		for(int i=0; i<8; i++) {
			timestamp[i] = array[pointer+i];
		}
		
		return timestamp;
	}    
    /**
     * Takes a byte array starting with a string padded with null
     * characters so that the length of the entire block is a multiple
     * of 4, and seperates it into a String and a byte array of the
     * remaining data.  These are then returned in a Vector.
     *
     * @param   block           block of data beginning with a string
     * @param   stringLength    number of characters in the string
    */
    static Vector<Object> getStringAndData(byte[] block, int stringLength) {
		Vector<Object> v = new Vector<Object>();
		int i;

		if ( stringLength %4 != 0) {
			System.out.println("getStringAndData: bad boundary");
			return v;
		}

		for (i = 0; block[i] != '\0'; i++) {
			if (i >= stringLength) {
				System.out.println("getStringAndData: Unreasonably long string");
				return v;
			}
		}
		// v.firstElement() is the String
		v.addElement(new String(Utils.copyBytes(block, 0, i)));

		i++;
		for (; (i % 4) != 0; i++) {
			if (i >= stringLength) {
				System.out.println("getStringAndData: Unreasonably long string");
				return v;
			}
			if (block[i] != '\0')
			{
				System.out.println("getStringAndData: Incorrectly padded string.");
				return v;
			}
		}
		// v.elementAt(1) is the position in the orginal byte[] where the data starts
		v.addElement( new Integer(i) );
		// v.lastElement() is the byte[] of data
		v.addElement( Utils.copyBytes( block, i, block.length - i) );
		return v;
	}
	/**
	 * Returns an array of Vectors containing types and arguments.
	 *
	 * @param   block   byte array containing types and arguments
	*/
	@SuppressWarnings("unchecked")
	static Vector<Object>[] getTypesAndArgs(byte[] block) {
		// TBD : throw exceptions or something when there are no type tags
		int n = block.length;
		@SuppressWarnings("rawtypes")
		Vector[] va = new Vector[2];
		if (n != 0) {
			if (block[0] == ',') {
				if (block[1] != ',') {
					/* This message begins with a type-tag string */
					va = getTypeTaggedArgs( block );
				} else {
					/* Double comma means an escaped real comma, not a
					 * type string */
					va = getHeuristicallyTypeGuessedArgs( block );
				}
			} else {
			va = getHeuristicallyTypeGuessedArgs( block );
			}
		}
		return va;
	}
	/**
	 * Returns Vectors containing the types and arguments from a
	 * type-tagged byte array
	 *
	 * @param   block   a byte array with type-tagged data
	*/
	@SuppressWarnings("unchecked")
	static Vector<Object>[] getTypeTaggedArgs( byte[] block ) {
		Vector<Character> typeVector = new Vector<Character>();
		Vector<Object> argVector = new Vector<Object>();
		int p = 0;

		/* seperate the block into the types byte array and the
		 * argument byte array*/
		Vector<Object> typesAndArgs = getStringAndData(block, block.length);
		byte[] args = (byte[]) typesAndArgs.lastElement();

		for (int thisType=1; block[thisType] != 0; thisType++) {
			switch (block[thisType]) {
				case '[' :
					typeVector.addElement(new Character('['));
					break;

				case ']' :
					typeVector.addElement(new Character(']'));
					break;

				case 'i': case 'm': case 'c':
					typeVector.addElement(new Character('i'));
					argVector.addElement(new Integer( Utils.toInt( Utils.copyBytes(args, p, 4) )) );
					p += 4;
					break;

				case 'r':
					typeVector.addElement(new Character('r'));
					byte[] colorBytes = Utils.copyBytes(args, p, 4);
					int r = (int) colorBytes[0] & 0xFF;
					int g = (int) colorBytes[1] & 0xFF;
					int b = (int) colorBytes[2] & 0xFF;
					int a = (int) colorBytes[3] & 0xFF;
					argVector.addElement(new Color(r, g, b, a));
					p += 8;
					break;					
					
				case 'f':
					typeVector.addElement(new Character('f'));
					argVector.addElement(new Float( Utils.toFloat( Utils.copyBytes(args, p, 4) )) );
					p += 4;
					break;

				case 'h': case 't':
					typeVector.addElement(new Character('h'));
					argVector.addElement(new Long( Utils.toLong( Utils.copyBytes(args, p, 8) )) );
					p += 8;
					break;

				case 'd':
					typeVector.addElement(new Character('d'));
					argVector.addElement(new Double( Utils.toDouble( Utils.copyBytes(args, p, 8) )) );
					p += 8;
					break;

				case 's': case 'S':
					typeVector.addElement(new Character('s'));
					byte[] remaining = Utils.copyBytes(args, p, args.length - p);
					Vector<Object> v = getStringAndData( remaining, remaining.length );
					argVector.addElement(v.firstElement());
					p += ((Integer)v.elementAt(1)).intValue();
					break;

				case 'T':
					typeVector.addElement(new Character('T'));
					break;
				case 'F':
					typeVector.addElement(new Character('F'));
					break;
				case 'N':
					typeVector.addElement(new Character('N'));
					break;
				case 'I':
					typeVector.addElement(new Character('I'));
					break;

				default:
					System.out.println( "[Unrecognized type tag " + block[thisType] + "]" );
			}
		}

		@SuppressWarnings("rawtypes")
		Vector[] returnValue = new Vector[2];
		returnValue[0] = typeVector;
		returnValue[1] = argVector;
		return returnValue;
    }
    /**
     * Returns the arguments from a non-type-tagged byte array
     *
     * @param   block   a byte array containing data
    */
    static Vector<Object>[] getHeuristicallyTypeGuessedArgs( byte[] block ) {
		// TODO : handle packets without type tags
    	System.out.println("Bad OSC packet: No type tags");
		return null;
    }
 
	static String timeFormat(long time) {
		if (time == 0) return "-immediate-";
		DateFormat formatter = new SimpleDateFormat("yyyyMMdd - HH:mm:ss:SSS");
		return formatter.format(time);		
	} 
}